import pickle
import torch
from elsciRL.agents.agent_abstract import QLearningAgent
import gymnasium as gym
from stable_baselines3 import A2C
from PIL import Image # Used to generate GIF


class SB_A2C(QLearningAgent):
    def __init__(self, policy:str='MlpPolicy', env:gym.Env = None):
        self.epsilon: int = 0 # Not used currently but required for compatibility
        self.device = "auto" if torch.cuda.is_available() else "cpu" # A2C is meant to be run primarily on the CPU, especially when you are not using a CNN.
        self.a2c = A2C("MlpPolicy", env, verbose=0, device="cpu", learning_rate=0.0007, n_steps=500)
        if torch.cuda.is_available():
            print("---- A2C is meant to be run primarily on the CPU ----")
            print("Device:", "CPU")

    def policy(self, state: any) -> str:
        return self.a2c.predict(state)

    def learn(self, total_steps:int=100) -> float:
        self.a2c.learn(total_timesteps=total_steps)
    
    def test(self, env, render:bool=False):
        #mean_reward, std_reward = evaluate_policy(self.a2c, env, n_eval_episodes=1)
        # Using environment from agent may limit episodes based on prior experience
        #vec_env = self.a2c.get_env()
        
        vec_env = env
        obs, info = vec_env.reset()
        
        actions = []
        states = []

        done = False
        episode_reward = 0
        render_stack = []
        if render:
            render_stack.append(Image.fromarray(vec_env.render().astype('uint8')))
        while not done: 
            action, _state = self.a2c.predict(obs, deterministic=True)
            actions.append(int(action))
            obs, r, done, truncated, info = vec_env.step(action)
            episode_reward += r
            if render:
                render_stack.append(Image.fromarray(vec_env.render().astype('uint8')))
        
            #states.append(info[0]['obs'])
            states.append(info['obs'])
            #vec_env.render("human")

        #episode_reward = info[0]['episode']['r']
        if episode_reward > 0.5:
            print("----> ", episode_reward)
        
        return episode_reward, actions, states, render_stack
    
    def q_result(self):
        results = [0,0]
        total_q = results[0]
        mean_q = results[1]
        return total_q, mean_q
    
    def clone(self):
        clone = pickle.loads(pickle.dumps(self))
        return clone
    
    


import torch
from typing import List
from torch import Tensor

from elsciRL.encoders.encoder_abstract import StateEncoder

class ObjectEncoder():
    def __init__(self, local_objects):
        """Encoder for default state representation produced by the environment/engine."""
        self.local_objects = {obj: i for i, obj in enumerate(local_objects)}
        device = "cuda" if torch.cuda.is_available() else "cpu" # Make this optional choice with parameter
        self.vectors: Tensor = torch.cat([torch.eye(len(self.local_objects)), torch.zeros(1, len(self.local_objects))]).to(device)         # tensor needs to be defined to len(local_object)
    
    def encode(self, state:list = None, legal_actions:list = None, episode_action_history:list = None,
               indexed: bool = False) -> Tensor:
        """ NO CHANGE - Board itself is used as state as is and simply converted to a vector"""
        # Goes through every item in state and labels based on the known objects available in the environment
        # New vector encoded form, for Chess: 64x12 flattened into 768x1 int vector to denote object occurance
        # NOT BINARY vector, value is the occurance of each object type. 
        #  -> In chess this happens to be [1 or 0] because you cant have more than one piece in each position.
        state_encoded: Tensor = torch.tensor([self.local_objects.get(state_pos, len(self.local_objects)) for state_pos in state], 
                                             device=self.vectors.device)
        
        if (not indexed):
            state_encoded = self.vectors[state_encoded].flatten()

        return state_encoded    


import random
from elsciRL.agents.agent_abstract import Agent
import torch
from torch import Tensor

class RandomAgent(Agent):
    """This is simply a random decision maker, does not learn."""
    def __init__(self):
        super().__init__()

    def policy(self, state: Tensor, legal_actions: list) -> str:
        action = random.choice(legal_actions)
        return action
    
    def learn(self, state: Tensor, next_state: Tensor, r_p: float, action_code: str) -> float:
        # Do nothing.
        return None
    
    def q_result(self):
        """Random agent has no knowledge."""
        total_q = 0
        mean_q = 0
        return total_q, mean_q



import numpy as np
import pandas as pd
import scipy.stats as st
# Graphs
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("Agg") # fixes error 'failed to allocate bitmap'
import seaborn as sns

class VisualOutput:
    def __init__(self, results_data, save_dir, show_figures, window_size):
        self.results_data = results_data
        self.save_dir = save_dir
        self.show_figures = show_figures
        self.num_episodes = np.max(results_data['episode'].max())
        self.agent_type = results_data['agent'][0].split('_')[0]
        if results_data['opponent'][0] is not None:
            self.opponent_type = results_data['opponent'][0].split('_')[0] # Not currently used but could be for titles?

        self.window_size = window_size

    def episode_reward_graph(self):
        # Plot RL Reward results for each approach
        # 1.1 Summary of total REWARD
        x =  self.results_data['episode']
        y = self.results_data['episode_reward']
        fig, ax1 = plt.subplots()
        ax1.plot(x,y,'r', label=self.agent_type)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Reward of Agent by Episode", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/episode_R.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/episode_R.png', dpi=fig.dpi)
            plt.close()

    def cumulative_reward_graph(self):
        # Plot RL Reward results for each approach
        # 1.1 Summary of total REWARD
        x =  self.results_data['episode']
        y = self.results_data['cumulative_reward']
        fig, ax1 = plt.subplots()
        ax1.plot(x,y,'r', label=self.agent_type)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Cumulative Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Cumulative Reward of Agent by Episode", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/cumulative_R.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/cumulative_R.png', dpi=fig.dpi)
            plt.close()

    def average_reward_graph(self):
        # Plot RL avg in Reward by episode for each approach
        # 1.3 Summary of avg in REWARD
        reward_delta = []
        ci_L_lst = []
        ci_U_lst = []
        for n,r in enumerate(self.results_data['episode_reward']): # End when window size no longer fits
            # At start number of prior episodes is less than window size
            window_reward = np.array(self.results_data['episode_reward'][:n])
            
            mean = np.mean(window_reward)
            if len(window_reward) <=5:
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(st.sem(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(np.mean(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            # CI produces error if all values the same
            elif np.count_nonzero(window_reward == window_reward[0]) == len(window_reward):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            else:
                ci_L, ci_U = st.t.interval(confidence=0.95, df=len(window_reward)-1, loc=np.mean(window_reward), scale=st.sem(window_reward))

            reward_delta.append(mean)
            ci_L_lst.append(ci_L)
            ci_U_lst.append(ci_U)
    
        x =  self.results_data['episode']
        y = np.array(reward_delta)
        ci_L_lst = np.array(ci_L_lst)
        ci_U_lst = np.array(ci_U_lst)
        fig, ax1 = plt.subplots()
        ax1.plot(x,y,'r', label=self.agent_type)
        ax1.fill_between(x,ci_L_lst, ci_U_lst, color='r', alpha = 0.25)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Average Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Average Reward over all episodes \n with 95% Confidence Interval", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        fig.tight_layout()
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/avg_R.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/avg_R.png', dpi=fig.dpi)
            plt.close()

    def rolling_average_reward_graph(self):
        # Plot RL avg in Reward by episode for each approach
        # 1.3 Summary of avg in REWARD
        
        # Calculate delta as difference between reward obtained from prior episode
        # Segment delta into batches and average to 'smooth' results calculation        
        if (int(self.num_episodes*self.window_size)+1)<100:
            w = int(self.num_episodes*self.window_size)+1
        else:
            w = 100
        reward_delta = []
        ci_L_lst = []
        ci_U_lst = []
        for n,r in enumerate(self.results_data['episode_reward']): # End when window size no longer fits
            # At start number of prior episodes is less than window size
            if n<w:
                window_reward = np.array(self.results_data['episode_reward'][:n])
            else:
                window_reward = np.array(self.results_data['episode_reward'][n-w:n])
            
            mean = np.mean(window_reward)
            if len(window_reward) <=5:
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(st.sem(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(np.mean(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            # CI produces error if all values the same
            elif np.count_nonzero(window_reward == window_reward[0]) == len(window_reward):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            else:
                ci_L, ci_U = st.t.interval(confidence=0.95, df=len(window_reward)-1, loc=np.mean(window_reward), scale=st.sem(window_reward))

            reward_delta.append(mean)
            ci_L_lst.append(ci_L)
            ci_U_lst.append(ci_U)
    
        x =  self.results_data['episode']
        y = np.array(reward_delta)
        ci_L_lst = np.array(ci_L_lst)
        ci_U_lst = np.array(ci_U_lst)
        fig, ax1 = plt.subplots()
        ax1.plot(x,y,'r', label=self.agent_type)
        ax1.fill_between(x,ci_L_lst, ci_U_lst, color='r', alpha = 0.25)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Average Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Rolling Average Reward for window sizes of "+str(w)+" episodes \n with 95% Confidence Interval", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        fig.tight_layout()
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/rolling_avg_R.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/rolling_avg_R.png', dpi=fig.dpi)
            plt.close()
            
    def avg_return_dist_graph(self):
        if int(self.num_episodes*self.window_size) <100:
            w = int(self.num_episodes*self.window_size)
        else:
            w = 100
        reward_delta = []
        for n,r in enumerate(self.results_data['episode_reward']): # End when window size no longer fits
            # At start number of prior episodes is less than window size
            if n<w:
                window_reward = np.mean(self.results_data['episode_reward'][:n])
            else:
                window_reward = np.mean(self.results_data['episode_reward'][n-w:n])
            reward_delta.append(window_reward)
        r_df = pd.DataFrame({'rolling_avg_R':reward_delta})

        sns.ecdfplot(data=r_df, x="rolling_avg_R", color='r', label=self.agent_type)
        #sns.displot(data=self.results_data, x="episode_reward", kind="ecdf", label=self.agent_type)
        plt.title("CDF of the Rolling Avg. Reward per Episode of Each Agent \n Window sizes of "+str(w)+" episodes")
        plt.xlabel("Reward per Episode")
        plt.ylabel("Cumulative Probability")
        plt.legend(fancybox=True, shadow=True, framealpha=1)
        plt.tight_layout()
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            plt.savefig(self.save_dir+'/cum_dist_avg_R.png', bbox_inches='tight')
            plt.close()
        else:
            plt.savefig(self.save_dir+'/cum_dist_avg_R.png', bbox_inches='tight')
            plt.close()
    
    def episode_reward_dist_graph(self):
        # Plot RL Reward results for each approach
        # 1.2 Summary of REWARD per episode
        sns.displot(self.results_data['episode_reward'], label=self.agent_type,color='r')
        plt.title("Distribution of the Reward Obtained per Episode of Each Agent")
        plt.xlabel("Reward per Episode")
        plt.ylabel("Occurrence")
        plt.legend(fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            plt.savefig(self.save_dir+'/episode_R_dist.png', bbox_inches='tight')
            plt.close()
        else:
            plt.savefig(self.save_dir+'/episode_R_dist.png', bbox_inches='tight')
            plt.close()

    def number_actions_graph(self):
        #2.1 Number of actions
        x =  self.results_data['episode']
        y = self.results_data['num_actions']
        fig, ax1 = plt.subplots()
        ax1.scatter(x,y,c='b', label=self.agent_type)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Number Actions')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Number of Actions of Agent per Episode", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/number_actions.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/number_actions.png', dpi=fig.dpi)
            plt.close()

    def number_actions_dist_graph(self):    
        # 2.2 Plot distr of actions taken per game
        sns.displot(self.results_data['num_actions'], label=self.agent_type,color='b')
        plt.title("Distribution of the Number of Actions Taken per Episode")
        plt.xlabel("Num Actions per Episode")
        plt.ylabel("Occurrence")
        plt.legend(fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            plt.savefig(self.save_dir+'/number_actions_dist.png', bbox_inches='tight')
            plt.close()
        else:
            plt.savefig(self.save_dir+'/number_actions_dist.png', bbox_inches='tight')
            plt.close()

    def runtime_per_episode_graph(self):
        # 3.1 Run time of training agent
        x = self.results_data['episode']
        y1 = self.results_data['time_per_episode']
        fig, ax1 = plt.subplots()
        ax1.plot(x,y1,'k', label=self.agent_type)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Time (seconds)')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        plt.title("Average Runtime per Episode", y=1.08)
        ax1.legend(loc=2, bbox_to_anchor=(0.8, 1.1), fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/time_per_episode.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/time_per_episode.png', dpi=fig.dpi)
            plt.close()

    def runtime_dist_graph(self):    
        # 3.2 Plot distr of time taken per game
        sns.displot(self.results_data['time_per_episode'], label=self.agent_type,color='k')
        plt.title("Distribution of the Time Taken per Episode")
        plt.xlabel("Time per Episode")
        plt.ylabel("Occurrence")
        plt.legend(fancybox=True, shadow=True, framealpha=1)
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            plt.savefig(self.save_dir+'/time_per_episode_dist.png', bbox_inches='tight')
            plt.close()
        else:
            plt.savefig(self.save_dir+'/time_per_episode_dist.png', bbox_inches='tight')
            plt.close()        

    @staticmethod
    def combined_variance_analysis_graph(variance_results, save_dir:str, show_figures:str):
        
        cycol = 'brgcmyk'
        line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot']
        col = 0
        fig, axs = plt.subplots(2,2)
        for results in variance_results:
            num_episode = np.max(results['episode'])
            avg_r_mean_sorted = np.sort(results['avg_R_mean'])
            cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

            # Plot RL Reward results for each approach
            # 1.1 Summary of total REWARD
            c = cycol[col]
            l = line_styles[col]
            if col <= len(cycol):
                col+=1
            else:
                c = np.random.rand(len(x),3)
                l = 'solid'                
            x =  results['episode']
            avg_R = np.array(results['avg_R_mean'])
            avg_R_SE = np.array(results['avg_R_se'])
            cum_R = np.array(results['cum_R_mean'])
            cum_R_SE = np.array(results['cum_R_se'])
            time_mean = np.array(results['time_mean'])
            
            axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=str(results))
            axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
            axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
            axs[1,0].plot(x,cum_R, color=c, linestyle=l)
            axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
            axs[1,1].hist(time_mean, color=c, alpha=0.25)

        axs[0,0].set_xlabel("Episode")
        axs[0,0].set_ylabel('Reward')
        axs[0,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R")
        
        axs[0,1].set_ylabel("Cumulative Probability")
        axs[0,1].set_xlabel("Mean Reward per Episode Window")
        axs[0,1].set_title("CDF of Rolling Average R")
        
        axs[1,0].set_xlabel("Episode")
        axs[1,0].set_ylabel('Cumulative Reward')
        axs[1,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[1,0].set_title("Cumulative R with Std. Err.")
        
        axs[1,1].set_ylabel("Occurence")
        axs[1,1].set_xlabel("Time")
        axs[1,1].set_title("Dist of Time per Episode")

        #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
        #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
        fig.suptitle("Comparison of Agent Training Results")
        fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
        fig.set_size_inches(12, 8)
        fig.tight_layout()

        if show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(save_dir+'/training_variance_comparison.png', dpi=100)
            plt.close()
        else:
            fig.savefig(save_dir+'/training_variance_comparison.png', dpi=100)
            plt.close()


    def convergence_analysis_graph(self):
        

        # 1 - Reward graph
        if (int(self.num_episodes*self.window_size)+1)<100:
            w = int(self.num_episodes*self.window_size)+1
        else:
            w = 100
        reward_delta = []
        ci_L_lst = []
        ci_U_lst = []
        for n,r in enumerate(self.results_data['episode_reward']): # End when window size no longer fits
            # At start number of prior episodes is less than window size
            if n<w:
                window_reward = np.array(self.results_data['episode_reward'][:n])
            else:
                window_reward = np.array(self.results_data['episode_reward'][n-w:n])
            
            mean = np.mean(window_reward)
            if len(window_reward) <=5:
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(st.sem(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            elif np.isnan(np.mean(window_reward)):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            # CI produces error if all values the same
            elif np.count_nonzero(window_reward == window_reward[0]) == len(window_reward):
                ci_L=np.mean(window_reward)
                ci_U=np.mean(window_reward)
            else:
                ci_L, ci_U = st.t.interval(confidence=0.95, df=len(window_reward)-1, loc=np.mean(window_reward), scale=st.sem(window_reward))

            reward_delta.append(mean)
            ci_L_lst.append(ci_L)
            ci_U_lst.append(ci_U)
    
        x =  self.results_data['episode']
        y = np.array(reward_delta)
        ci_L_lst = np.array(ci_L_lst)
        ci_U_lst = np.array(ci_U_lst)
        
        fig, (ax1, ax2) = plt.subplots(1,2)
        fig.suptitle("Rolling Average Reward vs Average Q Values for Agent Training Results", y=0.98)
        fig.set_size_inches(12, 5)
        
        ax1.plot(x,y,'r')
        ax1.fill_between(x,ci_L_lst, ci_U_lst, color='r', alpha = 0.25)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Rolling Average Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        
        #2 - Average Q graph
        y_2 = self.results_data['q_mean']
        ax2.plot(x,y_2,'k-.')
        ax2.set_xlabel("Episode")
        ax2.set_ylabel('Average Q')
        ax2.axes.get_xaxis().set_ticks([0, self.num_episodes])
        
        #fig.legend(fancybox=True, shadow=True, framealpha=1)
        fig.tight_layout()
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/convergence_analysis.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/convergence_analysis.png', dpi=fig.dpi)
            plt.close()


        # 2 - Summary of total REWARD
        fig, (ax1, ax2) = plt.subplots(1,2)
        fig.suptitle("Cumulative Reward vs Average Q Values for Agent Training Results", y=0.98)
        fig.set_size_inches(12, 5)

        x =  self.results_data['episode']
        y = self.results_data['cumulative_reward']
        ax1.plot(x,y,'r', label=self.agent_type)
        ax1.set_xlabel("Episode")
        ax1.set_ylabel('Cumulative Reward')
        ax1.axes.get_xaxis().set_ticks([0, self.num_episodes])
        
        #2 - Average Q graph
        y_2 = self.results_data['q_mean']
        ax2.plot(x,y_2,'k-.')
        ax2.set_xlabel("Episode")
        ax2.set_ylabel('Average Q')
        ax2.axes.get_xaxis().set_ticks([0, self.num_episodes])
        
        #fig.legend(fancybox=True, shadow=True, framealpha=1)
        fig.tight_layout()
        if self.show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(self.save_dir+'/convergence_analysis_totalR.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(self.save_dir+'/convergence_analysis_totalR.png', dpi=fig.dpi)
            plt.close()


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os 


def combined_variance_analysis_graph(results_dir:str='', analysis_type='training', show_figures:str='N'):
    
    if results_dir == '':
        raise ValueError("Save directory not specified.")
    
    # Get sub-dir for each problem-experiment type
    problem_folders = [name for name in os.listdir(results_dir) if os.path.isdir(os.path.join(save_dir, name))]
    # Find experiment folders
    experiment_folders = {}
    for problem_dir in problem_folders:
        if problem_dir not in experiment_folders:
            experiment_folders[problem_dir] = {}
        # --- Get sub-dir for each experiment type
        exp_dir_list = []
        for name in os.listdir(results_dir+'/'+problem_dir):
            if os.path.isdir(os.path.join(results_dir+'/'+problem_dir, name)):
                exp_dir_list.append(name)
        experiment_folders[problem_dir] = exp_dir_list 
    # Combine results data
    variance_results = {}
    for problem_dir in experiment_folders:
        for experiment_dir in experiment_folders[problem_dir]:
            
            if analysis_type == 'training':
                experiment_name = problem_dir+'_'+experiment_dir+'_training'
                file_names = [name for name in os.listdir(results_dir+'/'+problem_dir+'/'+experiment_dir) if name[0:25] == 'training_variance_results']
            elif analysis_type == 'testing':
                experiment_name = problem_dir+'_'+experiment_dir+'_testing'
                file_names = [name for name in os.listdir(results_dir+'/'+problem_dir+'/'+experiment_dir) if name[0:24] == 'testing_variance_results']
            else:
                raise ValueError("Analysis type must be either 'training' or 'testing'.")

            if experiment_name not in variance_results.keys():
                variance_results[experiment_name] = {}

            for file in file_names:
                results = pd.read_csv(results_dir+'/'+problem_dir+'/'+experiment_dir+'/'+file)
                variance_results[experiment_name]['results'] = results
                variance_results[experiment_name]['agent'] = results['agent'].iloc[0]
                variance_results[experiment_name]['num_repeats'] = results['num_repeats'].iloc[0]

    cycol = 'brgcmyk'
    line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot']
    col = 0
    fig, axs = plt.subplots(2,2)
    for experiment in list(variance_results.keys()):
        results = variance_results[experiment]['results']
        num_episode = np.max(results['episode'])
        avg_r_mean_sorted = np.sort(results['avg_R_mean'])
        cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

        # Plot RL Reward results for each approach
        # 1.1 Summary of total REWARD
        c = cycol[col]
        l = line_styles[col]
        if col <= len(cycol):
            col+=1
        else:
            c = np.random.rand(len(x),3)
            l = 'solid'                
        x =  results['episode']
        avg_R = np.array(results['avg_R_mean'])
        avg_R_SE = np.array(results['avg_R_se'])
        cum_R = np.array(results['cum_R_mean'])
        cum_R_SE = np.array(results['cum_R_se'])
        time_mean = np.array(results['time_mean'])
        
        axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=str(experiment))
        axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
        axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
        axs[1,0].plot(x,cum_R, color=c, linestyle=l)
        axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
        axs[1,1].hist(time_mean, color=c, alpha=0.25)

    axs[0,0].set_xlabel("Episode")
    axs[0,0].set_ylabel('Reward')
    axs[0,0].axes.get_xaxis().set_ticks([0, num_episode])
    axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R epi)")
    
    axs[0,1].set_ylabel("Cumulative Probability")
    axs[0,1].set_xlabel("Mean Reward per Episode Window")
    axs[0,1].set_title("CDF of Rolling Average R")
    
    axs[1,0].set_xlabel("Episode")
    axs[1,0].set_ylabel('Cumulative Reward')
    axs[1,0].axes.get_xaxis().set_ticks([0, num_episode])
    axs[1,0].set_title("Cumulative R with Std. Err.")
    
    axs[1,1].set_ylabel("Occurence")
    axs[1,1].set_xlabel("Time")
    axs[1,1].set_title("Dist of Time per Episode")

    #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
    #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
    fig.suptitle("COMBINED VARIANCE ANALYSIS")
    fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
    fig.set_size_inches(12, 8)
    fig.tight_layout()

    if show_figures == 'Y':
        plt.show(block=False)
        plt.pause(5)
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.close()
    else:
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.close()




from typing import Callable, Any
import gymnasium as gym
from gymnasium.envs.registration import register
from gymnasium import spaces

class EngineToGym(gym.Env):
    def __init__(self):
        print("elsciRL Env transformed to Gym Env.")

    def load(self, Engine, engine_name:str=None, Adapter:Callable[[Any], Any]=None, setup_info:dict={}):
        self.engine = Engine(setup_info)
        self.Adapter = Adapter()
        self.reward_signal = None
        self.reward_signal_tracker = []
        #self.action_history = []
        #self.obs_history = []
        # Use name if given directly, otherwise check engine ledger
        if engine_name is not None:
            self.name = engine_name
        elif (self.engine.ledger['id'] != 'Unique Problem ID')&(self.engine.ledger['id'] != ''):
            self.name = self.engine.ledger['id']
        else:
            print("\n WARNING: Engine name not set, using default name --> set inside ledger [id] field.")
            self.name = "elsciRLEnv-v0"
            
        # --------------------------
        # Define observation and action spaces
        # - Observations are dictionaries with the agent's and the target's location.
        # - Each location is encoded as an element of {0, ..., `size`}^2, i.e. MultiDiscrete([size, size]).
        try:
            # First check if observation space is defined by the adapter
            self.observation_space = self.Adapter.observation_space
        except:
            # Then check if observation space is defined by the engine
            try:
                self.observation_space = self.engine.observation_space
            except AttributeError:
                # Otherwise, use default observation space
                print("WARNING: Observation space not defined in either adapter of engine.")

        # - A single dimension of N number of discrete actions 
        self.action_space = spaces.Discrete(self.engine.ledger['action_space_size'])
        # --------------------------
        self.render_mode = self.engine.ledger['render_data']['render_mode']

    def reset(self, seed=None, options=None):
        observation = self.engine.reset()
        obs_enc = self.Adapter.adapter(observation, self.engine.legal_move_generator(), 
                               self.engine.action_history, encode = True )
        obs_enc = obs_enc.cpu().numpy() # Need to get all on CPU before registering with GYM
        self.reward_signal_tracker = [] # Only give agent reward for first time it sees a sub-goal
        self.action_history = [] # Reset action history
        self.episode_reward = 0
        #self.obs_history = []
        return obs_enc, {}

    def step(self, action=None):
        # Gym step function combines elsciRL Engine step and Adapter
        observation, reward, terminated, info = self.engine.step(action=int(action))
        self.action_history.append(action)
        # if observation not in self.obs_history:
        #     reward += 0.05 # Give small reward to encourage exploration
        # self.obs_history.append(observation)
        info['obs'] = observation

        # Apply custom reward signal if defined
        # - Defined as dict:= {obs:reward, obs:reward, ...}
        if self.engine.reward_signal:
            if observation in self.engine.reward_signal:
                if observation not in self.reward_signal_tracker:
                    # Only override if new reward is higher
                    if self.engine.reward_signal[observation] > reward:
                        reward=self.engine.reward_signal[observation]
                    self.reward_signal_tracker.append(observation)
                    

        # If a language problem then we also want processed observation
        if self.engine.ledger['type'] =='Language':
            obs_adapted = self.Adapter.adapter(observation, self.engine.legal_move_generator(), 
                                    self.engine.action_history, encode = False)
            info['obs_adapted'] = obs_adapted
        obs_enc = self.Adapter.adapter(observation, self.engine.legal_move_generator(), 
                               self.engine.action_history, encode = True )
        obs_enc = obs_enc.cpu().numpy() # Need to get all on CPU before registering with GYM
        truncated = False
        self.episode_reward += reward
        return obs_enc, reward, terminated, truncated, info

    def render(self):
        return self.engine.render() 

    def close(self):
        self.engine.close()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self

    
@staticmethod
def GymRegistration(Engine, Adapter, setup_info:dict={}):
    """This provides a function for converting elsciRL engines into OpenAI Gym environments. \n
    elsciRL engines include a conditional action space which is not inherently supported by OpenAI Gym. \n
    Outputs Engine in the OpenAI Gym format with a wrapper for the elsciRL adapter.
    """
    # Translate Engine to OpenAI Gym class structure
    environment = EngineToGym()
    environment.load(Engine, 'Test-1', Adapter, setup_info)
    # Register and make the environment
    register(id=environment.name, entry_point=environment)
    gym_env = gym.make(environment.name)
    
    return gym_env



import torch
from torch import Tensor
import numpy.typing as npt
import random


class elsciRLInfo:
    def __init__(
        self,
        observed_states: dict | None = None,
        experience_sampling: dict | None = None,
        # tensor_index: dict | None = None,
    ) -> None:
        if not experience_sampling:
            self.experience_sampling = {}
        else:
            self.experience_sampling = experience_sampling

        if not observed_states:
            self.observed_states = {}
        else:
            self.observed_states = observed_states

    def observed_state_tracker(
        self,
        engine_observation: Tensor | npt.ArrayLike | list | None = None,
        language_state: Tensor | npt.ArrayLike | list | None = None,
    ):
        """Tracks adapted form of state from engine observation for unsupervised approaches."""
        if engine_observation not in self.observed_states:
            self.observed_states[engine_observation] = language_state

    def experience_sampling_add(
        self,
        engine_observation: Tensor | npt.ArrayLike | list | None = None,
        action: int | str | bool | None = None,
        next_observation: Tensor | npt.ArrayLike | list | None = None,
        reward: float = 0,
        terminated: bool = False,
    ):
        """Adds experience from interaction with the Live environment to sample from."""
        # --------------------------------------------------------------------------
        # Required if input observation is tensor as this cant be used for dict keys
        # - create tuple store transitions
        if type(engine_observation) is Tensor:
            engine_observation = tuple(engine_observation.cpu().numpy().flatten())
        if type(next_observation) is Tensor:
            next_observation = tuple(next_observation.cpu().numpy().flatten())
        # --------------------------------------------------------------------------
        # Get occurrence of current observation+action
        if engine_observation not in self.experience_sampling:
            self.experience_sampling[engine_observation] = {}
        if action not in self.experience_sampling[engine_observation]:
            self.experience_sampling[engine_observation][action] = {}
            self.experience_sampling[engine_observation][action]["obs_a_count"] = 1
        obs_a_count = (
            self.experience_sampling[engine_observation][action]["obs_a_count"] + 1
        )
        self.experience_sampling[engine_observation][action][
            "obs_a_count"
        ] = obs_a_count

        # Get occurrence of next obs given obs+action
        # - Compute prob, reward is static and set on first occurrence
        if next_observation in self.experience_sampling[engine_observation][action]:
            next_obs_count = (
                self.experience_sampling[engine_observation][action][next_observation][
                    "next_obs_count"
                ]
                + 1
            )
            prob = next_obs_count / obs_a_count
            self.experience_sampling[engine_observation][action][next_observation][
                "next_obs_count"
            ] = next_obs_count
            self.experience_sampling[engine_observation][action][next_observation][
                "prob"
            ] = prob
        else:
            self.experience_sampling[engine_observation][action][next_observation] = {}
            self.experience_sampling[engine_observation][action][next_observation][
                "next_obs_count"
            ] = 1
            self.experience_sampling[engine_observation][action][next_observation][
                "prob"
            ] = (1 / obs_a_count)
            self.experience_sampling[engine_observation][action][next_observation][
                "reward"
            ] = reward
            self.experience_sampling[engine_observation][action][next_observation][
                "terminated"
            ] = terminated

    def experience_sampling_legal_actions(
        self, engine_observation: Tensor | npt.ArrayLike | list | None = None
    ):
        """Returns a list of known actions from the experience."""
        if type(engine_observation) is Tensor:
            engine_observation = tuple(engine_observation.cpu().numpy().flatten())
            # state_tuple = tuple(engine_observation)
            # engine_observation = self.tensor_index.index(state_tuple)
        if engine_observation in self.experience_sampling:
            legal_actions = list(self.experience_sampling[engine_observation].keys())
        else:
            legal_actions = None
        return legal_actions

    def experience_sampling_step(
        self,
        engine_observation: Tensor | npt.ArrayLike | list | None = None,
        action: int | str | bool | None = None,
    ):
        """Outcome of action given current observation from sampled experience."""
        # If state-action has not been seen from live system
        engine_observation_shape = None
        next_obs = None
        if type(engine_observation) is Tensor:
            engine_observation_shape = engine_observation.shape
            engine_observation = tuple(engine_observation.cpu().numpy().flatten())

        if action not in self.experience_sampling[engine_observation]:
            next_obs = engine_observation
            reward = 0
            terminated = False
        # Select action from distribution of probabilities
        else:
            cumulative = 0
            rng = random.random()
            for next_obs in self.experience_sampling[engine_observation][action]:
                # first key is just the count of obs+action so skip over this
                if next_obs != "obs_a_count":
                    if (
                        self.experience_sampling[engine_observation][action][next_obs][
                            "prob"
                        ]
                        <= rng
                    ):
                        break
                    else:
                        cumulative += self.experience_sampling[engine_observation][
                            action
                        ][next_obs]["prob"]

            reward = self.experience_sampling[engine_observation][action][next_obs][
                "reward"
            ]
            terminated = self.experience_sampling[engine_observation][action][next_obs][
                "terminated"
            ]

        # --------------------------------------------------------------------------
        # Converts stored obs back from int to tensor to match env
        if (type(next_obs) is tuple) and (
            engine_observation_shape is not None
        ):  # Phil: had to fix AND if statements by separating fully
            next_obs = torch.tensor(next_obs).reshape(engine_observation_shape)
        # --------------------------------------------------------------------------
        return next_obs, reward, terminated



ExperimentConfigData = {
    "name": "Example Experiment",
    "problem_type": "Examples",
    
    "experience_sample_batch_ratio": 0,
    
    "number_training_episodes": 1000,
    "number_training_repeats": 5,
    "number_training_seeds": 1,

    "test_agent_type":"all",
    "number_test_episodes": 100,
    "number_test_repeats": 5,

    "agent_select": ["Qlearntab", "Qlearntab"],
    "agent_parameters":{
        "Qlearntab":{
            "alpha": 0.1,
            "gamma": 0.95,
            "epsilon": 0.2,
            "epsilon_step":0.01
            }
        }
    }


import numpy as np
import torch
from multiprocessing.spawn import import_main_path
from typing import Dict, List, Tuple
from collections import Counter
from gymnasium.spaces import Box

from torch import Tensor
from elsciRL.encoders.encoder_abstract import StateEncoder

# Language Encoder
from sentence_transformers import SentenceTransformer

class LanguageEncoder(StateEncoder):
    """TODO: Specify required package to download for this encoder.
    Not included in main package requirements as it is a large download."""
    _cached_enc: Dict[str, Tensor] = dict()
    _cached_freq: Counter = Counter()

    def __init__(self, device: str = None):
        autodev = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device if device else autodev
        self.sentence_model: SentenceTransformer = SentenceTransformer('all-MiniLM-L6-v2', device=self.device)
        low_array = [-1 for i in range(384)]
        high_array = [1 for i in range(384)]
        self.observation_space = Box(low=np.array(low_array), high=np.array(high_array), dtype=np.float32)

    def encode(self, state: List[str], legal_actions:list = None, episode_action_history:list = None, 
               indexed: bool = False) -> Tensor:
        
        # I think typing is overriding the input type anyway -> need to ensure sentences are split up
        if type(state) == type(''):
            state = state.split(".") 
            state = [s for s in state if s.strip()]
        elif (len(state) == 0):
            state = [""]
                
        to_encode = [sent for sent in state if sent not in LanguageEncoder._cached_enc]
        if (to_encode):
            encoded = self.sentence_model.encode(to_encode, convert_to_tensor=True, show_progress_bar = False)
            LanguageEncoder._cached_enc.update({to_encode[i]: encoded[i] for i in range(len(to_encode))})
        
        LanguageEncoder._cached_freq.update(state)
        LanguageEncoder._cached_freq.subtract(LanguageEncoder._cached_freq.keys())
        state_encoded = torch.stack([LanguageEncoder._cached_enc[sent] for sent in state])

        if (len(LanguageEncoder._cached_freq) > 10000):
            for key, freq in list(reversed(LanguageEncoder._cached_freq.most_common()))[:2000]:
                del LanguageEncoder._cached_enc[key]
                del LanguageEncoder._cached_freq[key]

        return state_encoded


import os
import json
import logging
# Define Agent's parameters for problem
# Opponent is considered a 'local' specification as benchamarks vary between setting

class LocalConfig:
    def __init__(self, config_file_path: str):
        if (config_file_path):
            with open(config_file_path) as config_file:
                self.data = json.load(config_file)
                self.config_file_path = config_file_path
                
        else:
            self.data = dict()
            self.config_path = ""
            logging.info("No arguments given, using default configuration...")
        
    def __getitem__(self, key: str):
        item = None

        if (key in self.__dict__):
            item = self.__dict__[key]
        else:
            item = self.data[key]

        return item

#TODO this is not universal at all !!!
class ProblemConfig(LocalConfig):
    """Local Config is used to define any problem specific parameters."""
    def __init__(self, config_path: str):
        super(ProblemConfig, self).__init__(config_path)
        # State form
        self.adapter_select = self.data.get("adapter_select", [""])
        # Enabled agent to be trained against multiple opponents in order provided
        self.training_opponent_agent = self.data.get("training_opponent_agent", "")
        self.testing_opponent_agent = self.data.get("testing_opponent_agent", "")
        
        self.training_setup = self.data.get("training_setup",'default')
        self.testing_setup = self.data.get("testing_setup",'default')
        
        self.training_action_cap = self.data.get("training_action_cap",1000) # Arbitrary number to ensure games dont last forever
        self.testing_action_cap = self.data.get("testing_action_cap",1000) # Arbitrary number to ensure games dont last forever
        # Reward Signal, should be consistent between all agent being compared
        self.reward_signal = self.data.get("reward_signal",[1,-0.1,0,0] )# [Value of winning, Value for draw, Value for each action, Value for reaching new state]
        # Sub-Goal Defined
        self.sub_goal = self.data.get("sub_goal",None)
        
class ConfigSetup(LocalConfig):
    def __init__(self, config_dir: str):
        super(ConfigSetup, self).__init__(config_dir)
        self.state_configs = ProblemConfig(os.path.join(config_dir))


from abc import ABC, abstractmethod
from typing import List, Any

class StateAdapter(ABC):
    def __init__(self, raw_state):
        super().__init__()
        # Define the fields that describe the state features:
        self.state: list = self._read(raw_state)

    @abstractmethod
    def _read(raw_state) -> list:
        # Read the data.
        # fill in the feature fields
        raise NotImplementedError
    
    def adapter(self):
        "Returns the adapted form, may require input flag for encoded or non-encoded output."
        
        
    def sample(self):
        """Returns a sample of an adapted state form (typically inital positiion of the environment)."""

        



# TODO: Simplify and remove sub-goals/elsciRL tracking
import time
import numpy as np
# ------ Imports -----------------------------------------
# Agent Setup
from elsciRL.environment_setup.imports import ImportHelper
# Evaluation standards
from elsciRL.environment_setup.results_table import ResultsTable
from elsciRL.environment_setup.elsciRL_info import elsciRLInfo


class Environment:
    """Interaction Loop for standard environments.
    REQUIRES:
        - Engine: Environment engine defined with elsciRLAI format
        - Adapters: Dictionary of local adapters with unique names: {"name_1": Adapter_1, "name_2": Adapter_2,...}
        - local_setup_info: Dictionary of local setup info (i.e. local config file)
    """
    def __init__(self, Engine, Adapters:dict, local_setup_info: dict):
        # --- INIT env from engine
        try:
            supervised_rewards = local_setup_info['data']['supervised_rewards']
            y_limit = local_setup_info['data']['y_limit']
            self.obs_precision = local_setup_info['data']['obs_precision']
        except:
            supervised_rewards = local_setup_info['supervised_rewards']
            y_limit = local_setup_info['y_limit']
            self.obs_precision = local_setup_info['obs_precision']
        self.env = Engine(supervised_rewards=supervised_rewards, 
                          y_limit=y_limit,
                          obs_precision=self.obs_precision)
        self.start_obs = self.env.reset()
        # ---
        # --- PRESET elsciRL INFO
        # Agent
        Imports = ImportHelper(local_setup_info)
        self.agent, self.agent_type, self.agent_name, self.agent_state_adapter = Imports.agent_info(STATE_ADAPTER_TYPES)
        self.num_train_episodes, self.num_test_episodes, self.training_action_cap, self.testing_action_cap, self.reward_signal = Imports.parameter_info()  
        # Training or testing phase flag
        self.train = Imports.training_flag()
        # --- elsciRL
        self.live_env, self.observed_states, self.experience_sampling = Imports.live_env_flag()
        # Results formatting
        self.results = ResultsTable(local_setup_info)
        # elsciRL input function
        # - We only want to init trackers on first batch otherwise it resets knowledge
        self.elsciRL = elsciRLInfo(self.observed_states, self.experience_sampling)
        # Env start position for instr input
        # Enable sub-goals
        if (local_setup_info['sub_goal'] is not None) & (local_setup_info['sub_goal']!=["None"]) & (local_setup_info['sub_goal']!="None"):
            self.sub_goal:list = local_setup_info['sub_goal']
        else:
            self.sub_goal:list = None
        # New: multi sub-goal paths used for continuous instruction plans
        # - Would not be defined by config files but rather experiment module so dont need to define in terms of local_setup_info
        self.multi_sub_goal:dict = None

    def episode_loop(self):
        # Mode selection (already initialized)
        if self.train:
            number_episodes = self.num_train_episodes
        else:
            number_episodes = self.num_test_episodes

        if self.sub_goal:
            self.sub_goal = sub_goal_precision(self.sub_goal, self.obs_precision)

        sub_goal_tracker = {} # Used to track sub-goal completion -> re-call most common end state for start of next with EXACT chaining
        sub_goal_best = 0
        self.sub_goal_end = None
        for episode in range(0, number_episodes):
            action_history = []
            # ---
            # Start observation is used instead of .reset() fn so that this can be overridden for repeat analysis from the same start pos
            obs = self.env.reset(start_obs=self.start_obs)
            if episode>0:
                if obs != "{n:.{d}f}".format(n=0, d=self.obs_precision)+'_'+"{:0.1f}".format(0):
                    print("\n ---------- ")
                    print("Start Observation: ", obs)
            legal_moves = self.env.legal_move_generator(obs)
            state = self.agent_state_adapter.adapter(state=obs, legal_moves=legal_moves, episode_action_history=action_history, encode=True)
            # ---
            start_time = time.time()
            episode_reward:int = 0
            # ---
            sub_goal_reward_obtained = False
            multi_sub_goal_tracker = [] # Used to tracked known sub-goal completion and reduce print clutter
            multi_goal_reward_obtained = []
            for action in range(0,self.training_action_cap):
                if self.live_env:
                    # Agent takes action
                    legal_moves = self.env.legal_move_generator(obs)
                    agent_action = self.agent.policy(state, legal_moves)
                    action_history.append(agent_action)
                    
                    next_obs, reward, terminated = self.env.step(state=obs, action=agent_action)
                    # Can override reward per action with small negative punishment
                    if reward==0:
                        reward = self.reward_signal[1]
                    
                    legal_moves = self.env.legal_move_generator(next_obs) 
                    next_state = self.agent_state_adapter.adapter(state=next_obs, legal_moves=legal_moves, episode_action_history=action_history, encode=True)
                    # elsciRL trackers    
                    self.elsciRL.observed_state_tracker(engine_observation=next_obs,
                                                        language_state=self.agent_state_adapter.adapter(state=next_obs, legal_moves=legal_moves, episode_action_history=action_history, encode=False))
                    
                    # MUST COME BEFORE SUB-GOAL CHECK OR 'TERMINAL STATES' WILL BE FALSE
                    self.elsciRL.experience_sampling_add(state, agent_action, next_state, reward, terminated)
                    # Trigger end on sub-goal if defined
                    if self.sub_goal:
                        if (type(self.sub_goal)==type(''))|(type(self.sub_goal)==type(0)):
                            if next_obs == self.sub_goal:
                                # Sub-goal no longer terminates episode but only gets reward once
                                if sub_goal_reward_obtained is False:
                                    reward = self.reward_signal[0]
                                    sub_goal_reward_obtained = True
                                else:
                                    reward = 0
                                #terminated = True
                                # Required if we want exact sub-goal end position for next instruction
                                if next_obs not in sub_goal_tracker:
                                    sub_goal_tracker[next_obs] = 1
                                else:
                                    sub_goal_tracker[next_obs] = sub_goal_tracker[next_obs] + 1
                                if sub_goal_tracker[next_obs] >= sub_goal_best:
                                    self.sub_goal_end = next_obs
                                    sub_goal_best = sub_goal_tracker[next_obs]
                                
                        elif (type(self.sub_goal)==type(list('')))|(type(self.sub_goal)==type([0])):    
                            if next_obs in self.sub_goal:
                                # Sub-goal no longer terminates episode but only gets reward once
                                if sub_goal_reward_obtained is False:
                                    reward = self.reward_signal[0]
                                    sub_goal_reward_obtained = True
                                else:
                                    reward = 0
                                #terminated = True
                                # Required if we want exact sub-goal end position for next instruction
                                if next_obs not in sub_goal_tracker:
                                    sub_goal_tracker[next_obs] = 1
                                else:
                                    sub_goal_tracker[next_obs] = sub_goal_tracker[next_obs] + 1
                                if sub_goal_tracker[next_obs] >= sub_goal_best:
                                    self.sub_goal_end = next_obs
                                    sub_goal_best = sub_goal_tracker[next_obs]
                                #print("---------------------------------")
                                #print("Sub-Goal Reached: ", next_obs)
                        else:
                            print("Sub-Goal Type ERROR: The input sub-goal type must be a str/int or list(str/int).") 
                    # New: multi sub-goal completion for continuous chaining
                    # - If sub-task completed, obtain scaled reward based on instruction scale
                    if self.multi_sub_goal:
                        for multi_goal in self.multi_sub_goal:
                            if next_obs in self.multi_sub_goal[multi_goal]['sub_goal']:
                                # Only get reward for first visit in episode
                                if multi_goal not in multi_goal_reward_obtained:
                                    multi_goal_reward_scale = self.multi_sub_goal[multi_goal]['reward_scale']
                                    reward = self.reward_signal[0]*multi_goal_reward_scale
                                if multi_goal not in multi_sub_goal_tracker:
                                    multi_sub_goal_tracker.append(multi_goal)
                                    print("\n - Multi sub-goal completed: ", multi_goal, " -> reward after scaling = ", reward)
                                break # Assume state is unique to completing one sub-step (first used)                   
                else:
                    # Experience Sampling
                    legal_moves = self.elsciRL.experience_sampling_legal_actions(state)
                    # Unknown state, have no experience to sample from so force break episode
                    if legal_moves == None:
                        break
                    
                    agent_action = self.agent.policy(state, legal_moves)
                    next_state, reward, terminated = self.elsciRL.experience_sampling_step(state, agent_action)

                if self.train:
                    self.agent.learn(state, next_state, reward, agent_action)
                episode_reward+=reward
                if terminated:
                    break
                else:    
                    state=next_state
                    if self.live_env:
                        obs = next_obs        
            # If action limit reached
            if not terminated:
                reward = self.reward_signal[2]     
                
            end_time = time.time()
            agent_results = self.agent.q_result()
            if self.live_env:
                self.results.results_per_episode(self.agent_name, None, episode, action, episode_reward, (end_time-start_time), action_history, agent_results[0], agent_results[1]) 
        if (self.sub_goal):
            if (sub_goal_best>0):
                print("\n \t - Most Common Sub-Goal Reached: ", self.sub_goal_end ,"\n ---------- ")
            else:
                print("\n \t - Sub-Goal Not Reached \n ---------- ")
        return self.results.results_table_format()
                    


# TODO: SIMPLIFY TO JUST BASE TRAIN/TEST PROTOCOL WITHOUT EXPERIENCE SAMPLING
import os
import json
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent
from elsciRL.agents.agent_abstract import Agent

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialised and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> DONE: This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalize naming for this

class Experiment:
    """This is the standard Reinforcement Learning experiment setup for a flat agent. 
    - The agent is trained for a fixed number of episodes
    - Then learning is fixed to be applied during testing phase
    - Repeats (or seeds if environment start position changes) are used for statistical significant testing
    - Experience Sampling stores observed episodes into a sampled MDP model to learn from to improve training efficiency
    """
    def __init__(self, Config:dict, ProblemConfig:dict, Engine, Adapters:dict, save_dir:str, show_figures:str, window_size:float): 
        # Environment setup
        # - Multiple Engine support
        if isinstance(Engine, dict):
            print("\n Multiple Engines detected, will compare results across engines...")
            self.engine_comparison = True
            self.engine_list = Engine
        else:
            self.engine_comparison = False
            self.engine_list = {'DefaultEng':Engine}
        self.adapters = Adapters
        self.env = StandardInteractionLoop 
        # ---
        # Configuration setup
        self.ExperimentConfig = Config
        self.LocalConfig = ProblemConfig
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        self.save_dir = save_dir+'/Standard_Experiment'
        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y')|(self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")

        try:
            self.setup_info = self.ExperimentConfig['data'] | self.LocalConfig['data'] 
        except:
            self.setup_info = self.ExperimentConfig | self.LocalConfig 
        self.training_setups: dict = {}
        # new - store agents cross training repeats for completing the same start-end goal
        self.trained_agents: dict = {}
        self.num_training_seeds = self.setup_info['number_training_seeds']
        # new - config input defines the re-use of trained agents for testing: 'best' or 'all'
        self.test_agent_type = self.setup_info['test_agent_type']
        self.analysis = Evaluation(window_size=window_size)

        # ---
        # Agent setup
        self.AGENT_TYPES = {
            "Qlearntab": TableQLearningAgent,
            "Neural_Q": NeuralQLearningAgent,
            "Neural_Q_2": NeuralQLearningAgent,
            "Neural_Q_language": NeuralQLearningAgent
        }
        
        self.PLAYER_PARAMS = {
            "Qlearntab": ["alpha", "gamma", "epsilon"],
            "Neural_Q": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
            "Neural_Q_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
            "Neural_Q_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"]
        }

    def add_agent(self, agent_name:str, agent):
        """Add a custom agent to the experiment using the agent name as a key.
            - Paramters must be defined in the config.json file with matching name."""
        self.AGENT_TYPES[agent_name] = agent
        print("\n Agent added to experiment, all available agents: ", self.AGENT_TYPES)

    def train(self):
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)
        
        for engine_name, engine in self.engine_list.items():
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                # We are adding then overriding some inputs from general configs for experimental setups
                train_setup_info = self.setup_info.copy()
                # TODO: REMOVE experience sampling from standard and move to separate experiment
                if train_setup_info['experience_sample_batch_ratio']>0:
                    print("NOTE - Experience Sampling feature not currently implemented and will not be used")
                    train_setup_info['experience_sample_batch_ratio'] = 0
                # ----- State Adapter Choice
                adapter = train_setup_info["adapter_select"][n]
                # ----- Agent parameters
                agent_parameters = train_setup_info["agent_parameters"][agent_type]
                train_setup_info['agent_type'] = agent_type
                train_setup_info['agent_name'] = str(engine_name) + str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
                train_setup_info['adapter_select'] = adapter
                # -----
                # Repeat training
                train_setup_info['train'] = True
                number_training_episodes = train_setup_info['number_training_episodes']
                number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                print("Training Agent " + str(agent_type) + " for " + str(number_training_repeats) + " repeats on " + str(engine_name) + " engine")
                if str(engine_name) + '_' + str(agent_type) + '_' + str(adapter) not in self.trained_agents:
                    self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)] = {}

                seed_recall = {}
                seed_results_connection = {}
                for seed_num in range(0,self.num_training_seeds):
                    if self.num_training_seeds > 1:
                        print("------")
                        print("- Seed Num: ", seed_num)
                    # -------------------------------------------------------------------------------
                    # Initialise Environment
                    # Environment now init here and called directly in experimental setup loop
                    # - NEW: need to pass start position from live env so that experience can be sampled
                    if seed_num==0:
                        train_setup_info['training_results'] = False
                        train_setup_info['observed_states'] = False
                        train_setup_info['experience_sampling'] = False
                    else:
                        train_setup_info['training_results'] = False
                        train_setup_info['observed_states'] = observed_states_stored.copy()
                        train_setup_info['experience_sampling'] = experience_sampling_stored.copy()
                    # ---
                    setup_num:int = 0
                    temp_agent_store:dict = {}
                    for training_repeat in range(1,number_training_repeats+1):
                        if number_training_repeats > 1:
                            print("------")
                            print("- Repeat Num: ", training_repeat)
                        setup_num+=1
                        
                        # ----- init agent
                        player = self.AGENT_TYPES[agent_type](**agent_parameters) 
                        train_setup_info['agent'] = player 
                        
                        train_setup_info['live_env'] = True
                        live_env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                        
                        if training_repeat > 1:
                            live_env.start_obs = env_start

                        env_start = live_env.start_obs
                        goal = str(env_start).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        if goal in seed_recall:
                            setup_num = seed_recall[goal]
                        else:
                            seed_recall[goal] = 1
                        # - Results save dir -> will override for same goal if seen in later seed
                        if self.num_training_seeds > 1:
                            agent_save_dir = self.save_dir+'/'+engine_name+'_'+agent_type+'_'+adapter+'__training_results_'+str(goal)+'_'+str(setup_num) 
                        else:
                            agent_save_dir = self.save_dir+'/'+engine_name+'_'+agent_type+'_'+adapter+'__training_results_'+str(setup_num)
                        if not os.path.exists(agent_save_dir):
                            os.mkdir(agent_save_dir)

                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)]:
                            live_env.agent = self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)][goal].clone()

                        if train_setup_info['experience_sample_batch_ratio']>0:
                            live_sample_batch_size = int(number_training_episodes*train_setup_info['experience_sample_batch_ratio'])
                            live_sample_batch_count = int(1/train_setup_info['experience_sample_batch_ratio'])
                            # Train on Live system for limited number of total episodes
                            live_env.num_train_episodes = live_sample_batch_size
                            print("-- Training with Simulated Batches, ", live_sample_batch_count, " total...")
                            # init simulated environment
                            train_setup_info['live_env'] = False
                            simulated_env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                            simulated_env.start_obs = env_start
                            # Connect live agent -> This should be linked so continues learning
                            simulated_env.agent = live_env.agent

                            for live_sample_batch in range(0, live_sample_batch_count-1):
                                print("--- Live Interaction Batch Num", live_sample_batch+1)                            
                                training_results = live_env.episode_loop()

                                # Train based on simulated exp  
                                simulated_env.num_train_episodes = number_training_episodes
                                simulated_env.episode_loop()
                            # Final batch doesn't require simulated exp after and we need result output
                            print("--- Final batch")
                            train_setup_info['live_env'] = True
                            train_setup_info['number_training_episodes'] = live_sample_batch_size
                            training_results = live_env.episode_loop()
                            # Have to 'fix' output
                            training_results['episode'] = training_results.index
                            cumulative_r = 0
                            cumulative_r_lst = []
                            for r in training_results['episode_reward']:
                                cumulative_r+=r
                                cumulative_r_lst.append(cumulative_r)
                            training_results['cumulative_reward'] = cumulative_r_lst
                        else:
                            # ---
                            if goal in seed_results_connection:
                                live_env.results.load(seed_results_connection[goal])
                            #live_env.agent.exploration_parameter_reset()
                            training_results = live_env.episode_loop()
                            training_results['episode'] = training_results.index
                        # Opponent now defined in local setup.py
                        # ----- Log training results      
                        training_results.insert(loc=0, column='Repeat', value=setup_num)
                        # Produce training report with Analysis.py
                        Return = self.analysis.train_report(training_results, agent_save_dir, self.show_figures)
                        # Extract trained agent from env and stored for re-call
                        if goal not in temp_agent_store:
                            temp_agent_store[goal] = {}
                        temp_agent_store[goal][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                        
                        if training_repeat == 1:
                            max_Return = Return
                            best_agent = live_env.agent
                            training_results_stored =  live_env.results.copy()
                            observed_states_stored = live_env.elsciRL.observed_states
                            experience_sampling_stored = live_env.elsciRL.experience_sampling
                        if Return > max_Return:
                            max_Return = Return
                            best_agent = live_env.agent
                            training_results_stored =  live_env.results.copy()
                            observed_states_stored = live_env.elsciRL.observed_states
                            experience_sampling_stored = live_env.elsciRL.experience_sampling
                        seed_recall[goal] = seed_recall[goal] + 1
                        # Save trained agent to logged output
                        train_setup_info['train_save_dir'] = agent_save_dir
                        #train_setup_info['trained_agent'] = agent
                    seed_results_connection[goal] = training_results_stored

                    # ----- New: 'best' or 'all' agents saved
                    # Save trained agent to logged output for testing phase
                    if self.test_agent_type.lower() == 'best':
                        self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)][goal] = best_agent.clone()
                    elif self.test_agent_type.lower() == 'all':
                        start_repeat_num = list(temp_agent_store[goal].keys())[0]
                        end_repeat_num = list(temp_agent_store[goal].keys())[-1]

                        all_agents = []
                        for repeat in range(start_repeat_num,end_repeat_num+1):
                            agent = temp_agent_store[goal][repeat]['agent']
                            all_agents.append(agent)
                            
                        if goal not in self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)]:
                            self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)][goal] = {}
                        self.trained_agents[str(engine_name) + '_' + str(agent_type) + '_' + str(adapter)][goal] = all_agents

                    # Collate complete setup info to full dict
                self.training_setups['Training_Setup_'+str(engine_name) + '_' + str(agent_type)+'_'+str(adapter)] = train_setup_info
        if (number_training_repeats>1)|(self.num_training_seeds):
            self.analysis.training_variance_report(self.save_dir, self.show_figures)
        #json.dump(self.training_setups) # TODO: Won't currently serialize this output to a json file
        return self.training_setups

    # TESTING PLAY
    def test(self, training_setups:str=None):
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
        else:
            training_setups = json.load(training_setups)

        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            test_setup_info['training_results'] = False
            test_setup_info['observed_states'] = False
            test_setup_info['experience_sampling'] = False
            print("----------")
            print("Testing results for trained agents in saved setup configuration:")
            print(test_setup_info['train_save_dir'])
            number_training_repeats = test_setup_info['number_test_repeats']
            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']

            # Only use the trained agent with best return
            if self.test_agent_type.lower()=='best':
                for engine_name, engine in self.engine_list.items():
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                        # Re-init env for testing
                        env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agent available for testing.")
                            env.agent = self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        env.agent.epsilon = 0 # Remove random actions
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        testing_results = env.episode_loop() 
                        test_save_dir = (self.save_dir+'/' + str(engine_name) + '_' + agent_adapter + '__testing_results_' + str(goal).split("/")[0]+"_"+str(testing_repeat))
                        if not os.path.exists(test_save_dir):
                            os.mkdir(test_save_dir)
                        # Produce training report with Analysis.py
                        Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)
                        
            # Re-apply all trained agents with fixed policy
            elif self.test_agent_type.lower()=='all':
                # All trained agents are used:
                # - Repeats can be used to vary start position
                # - But assumed environment is deterministic otherwise
                # Re-init env for testing
                for engine_name, engine in self.engine_list.items():
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):
                        env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agents available for testing.")
                            all_agents = self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        
                        for ag,agent in enumerate(all_agents):
                            env.results.reset() # Reset results table for each agent
                            env.start_obs = start_obs
                            env.agent = agent
                            env.agent.epsilon = 0 # Remove random actions
                            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                            # ---
                            # Testing generally is the agents replaying on the testing ENV
                            testing_results = env.episode_loop() 
                            test_save_dir = (self.save_dir+'/'+ str(engine_name) + '_' + agent_adapter + '__testing_results_' + str(goal).split("/")[0]+"_"+"agent"+str(ag)+"-repeat"+str(testing_repeat))
                            if not os.path.exists(test_save_dir):
                                os.mkdir(test_save_dir)
                            # Produce training report with Analysis.py
                            Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)

        if (number_training_repeats>1)|(self.test_agent_type.lower()=='all'):
            self.analysis.testing_variance_report(self.save_dir, self.show_figures)

    def render_results(self, training_setups:str=None):
        """Apply fixed policy to render current decision making for limited number of episodes."""
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
        else:
            json.load(training_setups)

        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            test_setup_info['training_results'] = False
            test_setup_info['observed_states'] = False
            test_setup_info['experience_sampling'] = False
            print("----------")
            print("Rendering trained agent's policy:")
            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']


            if self.test_agent_type.lower()=='best':
                for engine_name, engine in self.engine_list.items():
                    # Re-init env for testing
                    env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                    # ---
                    start_obs = env.start_obs
                    goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                    print("Flat agent Goal: ", goal)
                    # Override with trained agent if goal seen previously
                    if goal in self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                        print("Trained agent available for testing.")
                        env.agent = self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                    else:
                        print("NO agent available for testing position.")
                    env.agent.epsilon = 0 # Remove random actions
                    # ---
                    # Testing generally is the agents replaying on the testing ENV
                    render_save_dir = self.save_dir+'/render_results'
                    if not os.path.exists(render_save_dir):
                        os.mkdir(render_save_dir)
                    render_results = env.episode_loop(render=True, render_save_dir=render_save_dir) 
                    # Produce training report with Analysis.py
                    Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)
            else:  
                for engine_name, engine in self.engine_list.items():
                    ## Re-init env for testing
                    env = self.env(Engine=engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                    # ---
                    start_obs = env.start_obs
                    goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                    print("Flat agent Goal: ", goal)
                    # Override with trained agent if goal seen previously
                    if goal in self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                        print("Trained agents available for testing.")
                        all_agents = self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                    else:
                        print("NO agent available for testing position.")
                    #Only render first even if all selected
                    for ag,agent in enumerate(all_agents[:1]):
                        env.results.reset() # Reset results table for each agent
                        env.start_obs = start_obs
                        env.agent = agent
                        env.agent.epsilon = 0 # Remove random actions
                        agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        render_save_dir = self.save_dir+'/render_results'
                        if not os.path.exists(render_save_dir):
                            os.mkdir(render_save_dir)
                        render_results = env.episode_loop(render=True, render_save_dir=render_save_dir) 
                        
                        # Produce training report with Analysis.py
                        Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)


from .experiments.standard import Experiment as STANDARD_RL
from .instruction_following.supervised_instruction_following import SupervisedExperiment as SUPERVISED_RL_HIERARCHY
from .instruction_following.unsupervised_instruction_following import UnsupervisedSearch as UNSUPERVISED_RL_HIERARCHY
from .instruction_following.elsciRL_instruction_search import elsciRLSearch as elsciRL_SEARCH
from .instruction_following.elsciRL_instruction_following import elsciRLOptimize as elsciRL_OPTIMIZE

from .analysis.combined_variance_visual import combined_variance_analysis_graph as COMBINED_VARIANCE_ANALYSIS_GRAPH



import pandas as pd


class ResultsTable:
    def __init__(self, local_setup_info:dict = None) -> None:
        if type(local_setup_info['training_results']) != type(pd.DataFrame()):
            self.agent:list = []
            self.opponent:list =[]
            self.episode:list = []
            self.num_actions:list = []
            self.episode_reward:list = []
            self.cumulative_reward:list = []
            self.time_per_episode:list = []
            self.action_history:list = []
            self.q_total:list = []
            self.q_mean:list = []
            # new
            self.cum_r = 0
        else:
            self.agent:list = local_setup_info['training_results'].agent.tolist()
            self.opponent:list =local_setup_info['training_results'].opponent.tolist()
            self.episode:list = local_setup_info['training_results'].episode.tolist()
            self.num_actions:list = local_setup_info['training_results'].num_actions.tolist()
            self.episode_reward:list = local_setup_info['training_results'].episode_reward.tolist()
            self.cumulative_reward:list = local_setup_info['training_results'].cumulative_reward.tolist()
            self.cum_r = self.cumulative_reward[-1]
            self.time_per_episode:list = local_setup_info['training_results'].time_per_episode.tolist()
            self.action_history:list = local_setup_info['training_results'].action_history.tolist()
            self.q_total:list = local_setup_info['training_results'].q_total.tolist()
            self.q_mean:list = local_setup_info['training_results'].q_mean.tolist()

    def results_per_episode(self,agent_name:str='missing', opponent_name:str='None', episode_num:int=0, action_num:int=0, episode_reward:float=0, time:float=0, episode_action_history:list=[], q_total:float=0, q_mean:float=0):
        self.agent.append(agent_name)
        self.opponent.append(opponent_name)
        self.episode.append(episode_num)
        self.num_actions.append(action_num)
        self.episode_reward.append(episode_reward)
        self.cum_r +=episode_reward
        self.cumulative_reward.append(self.cum_r)
        self.time_per_episode.append(time)
        self.action_history.append(episode_action_history)
        self.q_total.append(q_total)
        self.q_mean.append(q_mean)
        

    def results_table_format(self):
        results= pd.DataFrame({ 
                    'agent': self.agent,
                    'opponent': self.opponent,
                    'episode': self.episode, 
                    'num_actions': self.num_actions, 
                    'episode_reward': self.episode_reward,
                    "cumulative_reward": self.cumulative_reward,
                    "time_per_episode":self.time_per_episode,
                    "action_history": self.action_history,
                    "q_total":self.q_total,
                    "q_mean":self.q_mean})
        return results
    
    def reset(self):
        self.agent:list = []
        self.opponent:list =[]
        self.episode:list = []
        self.num_actions:list = []
        self.episode_reward:list = []
        self.cum_r = 0
        self.cumulative_reward:list = []
        self.time_per_episode:list = []
        self.action_history:list = []
        self.q_total:list = []
        self.q_mean:list = []

    def copy(self):
        results_copy= pd.DataFrame({ 
                    'agent': self.agent.copy(),
                    'opponent': self.opponent.copy(),
                    'episode': self.episode.copy(), 
                    'num_actions': self.num_actions.copy(), 
                    'episode_reward': self.episode_reward.copy(),
                    "cumulative_reward": self.cumulative_reward.copy(),
                    "time_per_episode":self.time_per_episode.copy(),
                    "action_history":self.action_history.copy(),
                    "q_total":self.q_total.copy(),
                    "q_mean":self.q_mean.copy()})
        return results_copy
    
    def load(self, results_copy):
        self.agent:list = results_copy.agent.tolist()
        self.opponent:list = results_copy.opponent.tolist()
        self.episode:list = results_copy.episode.tolist()
        self.num_actions:list = results_copy.num_actions.tolist()
        self.episode_reward:list = results_copy.episode_reward.tolist()
        self.cumulative_reward:list = results_copy.cumulative_reward.tolist()
        self.cum_r = self.cumulative_reward[-1]
        self.time_per_episode:list = results_copy.time_per_episode.tolist()
        self.action_history:list = results_copy.action_history.tolist()
        self.q_total:list = results_copy.q_total.tolist()
        self.q_mean:list = results_copy.q_mean.tolist()





# TODO: Simplify and remove sub-goals/elsciRL tracking/live_env/exp sampling
import time
import numpy as np
from tqdm import tqdm
# ------ Imports -----------------------------------------
# Agent Setup
from elsciRL.environment_setup.imports import ImportHelper
# Evaluation standards
from elsciRL.environment_setup.results_table import ResultsTable
from elsciRL.environment_setup.elsciRL_info import elsciRLInfo
# Non-gym interaction loop setup
from elsciRL.interaction_loops.standard import StandardInteractionLoop

class GymInteractionLoop:
    """Interaction Loop for standard environments.
    REQUIRES:
        - Engine: Environment engine defined with elsciRLAI format
        - Adapters: Dictionary of local adapters with unique names: {"name_1": Adapter_1, "name_2": Adapter_2,...}
        - local_setup_info: Dictionary of local setup info (i.e. local config file)
    """
    def __init__(self, Engine, Adapters:dict, local_setup_info: dict):
        # Define agent type for interaction process, call alternative if not gym agent
        if local_setup_info['agent_type'].split('_')[0] == "SB3":
            self.gym_agent = True
            # --- INIT env from engine
            self.env = Engine(local_setup_info)
            self.start_obs = self.env.reset()
            # ---
            # --- PRESET elsciRL INFO
            # Agent
            Imports = ImportHelper(local_setup_info)
            self.agent, self.agent_type, self.agent_name, self.agent_state_adapter = Imports.agent_info(Adapters)
            self.num_train_episodes, self.num_test_episodes, self.training_action_cap, self.testing_action_cap, self.reward_signal = Imports.parameter_info()  
            # Training or testing phase flag
            self.train = Imports.training_flag()
            # --- elsciRL
            self.live_env, self.observed_states, self.experience_sampling = Imports.live_env_flag()
            # Results formatting
            self.results = ResultsTable(local_setup_info)
            # elsciRL input function
            # - We only want to init trackers on first batch otherwise it resets knowledge
            self.elsciRL = elsciRLInfo(self.observed_states, self.experience_sampling)
        else:
            # --- Used for initialisation default interaction loop as alternative
            self.gym_agent = False
            self.interaction = StandardInteractionLoop(Engine, Adapters, local_setup_info)
            self.start_obs = self.interaction.start_obs
            self.results = ResultsTable(local_setup_info)

    def episode_loop(self, render:bool=False, render_save_dir:str=None):
        if self.gym_agent:
            # Mode selection (already initialized)
            if self.train:
                number_episodes = self.num_train_episodes
            else:
                number_episodes = self.num_test_episodes
            
            episode_render = []
            print("\n Episode Interaction Loop: ")
            if self.train:
                for episode in tqdm(range(0, number_episodes)):
                    start_time = time.time()
                    # Can force the agent to train on a single episode
                    # Very time consuming to do this
                    self.agent.learn(total_steps=self.training_action_cap)
                    end_time = time.time()
                    reward, actions, states, render_stack = self.agent.test(self.env, render=render)
                    episode_render.append(render_stack)
                    self.results.results_per_episode(self.agent_name, None, episode, len(actions), 
                                                    reward, (end_time-start_time), actions, 0, 0)  
            else:
                for episode in tqdm(range(0, number_episodes)):
                    start_time = time.time()
                    # Evaluate fixed policy on single episode
                    reward, actions, states, render_stack = self.agent.test(self.env, render=render)
                    episode_render.append(render_stack)
                    end_time = time.time()
                    self.results.results_per_episode(self.agent_name, None, episode, len(actions), 
                                                    reward, (end_time-start_time), actions, 0, 0)
            table_results = self.results.results_table_format()
            # Output GIF image of all episode frames
            if render:
                render_stack[0].save(render_save_dir+'/render.gif', save_all=True, append_images=render_stack[1:], optimize=False, duration=200, loop=1)
        else:
            table_results = self.interaction.episode_loop()
            self.agent = self.interaction.agent
            self.results = self.interaction.results
            self.elsciRL = self.interaction.elsciRL

        return table_results



#  Define data through class function so it can be called within package
# Instead of using a .json file which is hard to load from local install
# NOTE: MAKE SURE TO TRUST REPOSITORIES BEFORE RUNNING CODE
# - Can set branch to specific commit to ensure no changes are made without knowledge
#   |-----> changed to commit id which is tied to branch and more stable
class Applications:
    def __init__(self):
        self.data ={
            "Sailing":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-Sailing",
                "commit_id": "e384424",
                "engine_filenames": "engine",
                "local_config_filenames": {"easy":"easy_river.json", "medium":"medium_river.json"},
                "adapter_filenames": {"default":"default","language":"language"}
                }
        }


class elsciRLInput:
    def __init__(self, description_lookup:dict=None):
        self.description_lookup = description_lookup
        # New: store descriptions provided so the user doesn't need to provide multiple times
        self.descriptions_stored:dict={}
        
    def user_input(self):
        instructions = []
        instruction_descriptions = []
        while True:
            instr = input("Please provide the current instruction... ([e/exit] to end path)")
            if (instr == "e")|(instr=="exit"):
                break
            
            if not self.description_lookup:
                if instr not in self.descriptions_stored:
                    description = input("Please provide a description of the instruction...")
                else:
                    print("Instruction description provided previously.")
                    description = self.descriptions_stored[instr]
            if description == "None":
                description = instr
            
            instructions.append(instr)
            instruction_descriptions.append(description)
            self.descriptions_stored[instr] = description


        return instructions, instruction_descriptions